use std::simd::num::SimdUint;
use std::simd::{u8x4, Simd};

#[inline]
pub const fn get_digit(b: u8) -> isize {
    // wrapping_sub('0' as u32) same as - 48 but less magical
    (b as isize).wrapping_sub('0' as isize)
}

#[inline]
pub fn temp(bytes: &[u8]) -> isize {
    let is_negative = bytes[0] == b'-';
    let as_decimal = match (is_negative, bytes.len()) {
        (true, 4) => get_digit(bytes[1]) * 10 + get_digit(bytes[3]),
        (true, 5) => get_digit(bytes[1]) * 100 + get_digit(bytes[2]) * 10 + get_digit(bytes[4]),
        (false, 3) => get_digit(bytes[0]) * 10 + get_digit(bytes[2]),
        (false, 4) => get_digit(bytes[0]) * 100 + get_digit(bytes[1]) * 10 + get_digit(bytes[3]),
        _x => panic!(
            "could not parse temp: is_negative = {is_negative}, length = {}",
            bytes.len()
        ),
    };
    if is_negative {
        -as_decimal
    } else {
        as_decimal
    }
}

// not really faster
#[inline]
pub fn temp_new(bytes: &[u8]) -> isize {
    let (characteristic, mantissa) = unsafe { bytes.split_at_unchecked(bytes.len() - 2) };
    let mut sum = (mantissa[1] as u32).wrapping_sub('0' as u32);
    let mut position = 10;
    let mut idx = characteristic.len();
    while idx > 0 {
        idx -= 1;
        if characteristic[idx] == b'-' {
            return -(sum as isize);
        }
        if characteristic[idx] == 0u8 {
            return sum as isize;
        }
        sum += (characteristic[idx] as u32).wrapping_sub('0' as u32) * position;
        position *= 10;
    }
    sum as isize
}

// slower
#[inline]
pub fn temp_simd(bytes: &[u8]) -> isize {
    let is_negative = bytes[0] == b'-';
    let bytes = if is_negative { &bytes[1..] } else { bytes };
    let len = bytes.len();
    let idxs = Simd::from_array([5, len.wrapping_sub(4), len - 3, len - 1]);
    let zeroes = u8x4::splat(b'0');
    let temp_simd = Simd::gather_or(bytes, idxs, zeroes);
    let subbed = temp_simd - zeroes;
    let subbed: Simd<isize, 4> = subbed.cast();
    let mul = Simd::from_array([0, 100, 10, 1]);
    let mulled = subbed * mul;
    let sum: isize = mulled.to_array().iter().sum();
    if is_negative {
        -sum
    } else {
        sum
    }
}

#[cfg(test)]
mod tests {
    use crate::utils::parse::temp_new;

    #[test]
    fn test_temp_new_max() {
        let temp_max = temp_new("99.9".as_bytes());
        assert_eq!(temp_max, 999);
    }

    #[test]
    fn test_temp_new_min() {
        let temp_min = temp_new("-99.9".as_bytes());
        assert_eq!(temp_min, -999);
    }

    #[test]
    fn test_temp_new_zero() {
        let temp_0 = temp_new("0.0".as_bytes());
        assert_eq!(temp_0, 0);
    }

    #[test]
    fn test_temp_new_pos() {
        let temp_10 = temp_new("9.9".as_bytes());
        assert_eq!(temp_10, 99);
    }

    #[test]
    fn test_temp_new_neg() {
        let temp_neg_10 = temp_new("-9.9".as_bytes());
        assert_eq!(temp_neg_10, -99);
    }
}
